from abc import abstractmethod
from typing import TypedDict

from bleak import BleakClient


class RadonEyeStatus(TypedDict):
    serial: str
    model: str
    version: str
    latest_bq_m3: int
    latest_pci_l: float
    day_avg_bq_m3: int
    day_avg_pci_l: float
    month_avg_bq_m3: int
    month_avg_pci_l: float
    peak_bq_m3: int
    peak_pci_l: float
    counts_current: int
    counts_previous: int
    counts_str: str
    uptime_minutes: int
    uptime_str: str


class RadonEyeHistoryPage(TypedDict):
    page_count: int
    page_no: int
    value_count: int
    values_bq_m3: list[int]
    values_pci_l: list[float]


class RadonEyeHistory(TypedDict):
    values_bq_m3: list[int]
    values_pci_l: list[float]


class RadonEyeInterfaceBase:
    @classmethod
    @abstractmethod
    def supports(cls, client: BleakClient) -> bool:
        raise NotImplementedError

    @abstractmethod
    async def status(self, client: BleakClient) -> RadonEyeStatus:
        raise NotImplementedError

    @abstractmethod
    async def history(self, client: BleakClient) -> RadonEyeHistory:
        raise NotImplementedError

    @abstractmethod
    async def beep(self, client: BleakClient) -> None:
        raise NotImplementedError



from bleak import BleakClient

from radoneye.interface import RadonEyeInterfaceBase
from radoneye.interface_v1 import RadonEyeAdapterV1
from radoneye.interface_v2 import RadonEyeAdapterV2


class RadonEyeInterfaceFactory:
    @classmethod
    def create(
        cls, client: BleakClient, status_read_timeout: float, history_read_timeout: float
    ) -> RadonEyeInterfaceBase:
        if RadonEyeAdapterV1.supports(client):
            return RadonEyeAdapterV1(
                status_read_timeout=status_read_timeout,
            )
        if RadonEyeAdapterV2.supports(client):
            return RadonEyeAdapterV2(
                status_read_timeout=status_read_timeout,
                history_read_timeout=history_read_timeout,
            )
        else:
            raise NotImplementedError("Unsupported device version")



import asyncio
import json
from argparse import ArgumentDefaultsHelpFormatter, ArgumentParser
from typing import Literal, NamedTuple

from radoneye.client import RadonEyeClient
from radoneye.scanner import RadonEyeScanner


class ListCommandArgs(NamedTuple):
    adapter: str | None
    timeout: int
    output: Literal["text", "json"]


async def cmd_list(args: ListCommandArgs):
    for dev in await RadonEyeScanner.discover(adapter=args.adapter, timeout=args.timeout):
        if args.output == "text":
            print(f"{dev.address}\t{dev.name}")
        else:
            print(json.dumps({"address": dev.address, "name": dev.name}, separators=(",", ":")))


class BeepCommandArgs(NamedTuple):
    adapter: str | None
    connect_timeout: int
    address: str


async def cmd_beep(args: BeepCommandArgs):
    async with RadonEyeClient(
        args.address,
        adapter=args.adapter,
        connect_timeout=args.connect_timeout,
    ) as client:
        print(f"Beeping on {args.address}")
        await client.beep()


class StatusCommandArgs(NamedTuple):
    adapter: str | None
    connect_timeout: int
    read_timeout: int
    output: Literal["text", "json"]
    address: str


async def cmd_status(args: StatusCommandArgs):
    async with RadonEyeClient(
        args.address,
        adapter=args.adapter,
        connect_timeout=args.connect_timeout,
        status_read_timeout=args.read_timeout,
    ) as client:
        status = await client.status()
        if args.output == "text":
            for key in sorted(status.keys()):
                print(f"{key} = {status.get(key)}")
        else:
            print(json.dumps(status, separators=(",", ":")))


class HistoryCommandArgs(NamedTuple):
    adapter: str | None
    connect_timeout: int
    read_timeout: int
    output: Literal["text", "json"]
    address: str


async def cmd_history(args: HistoryCommandArgs):
    async with RadonEyeClient(
        args.address,
        adapter=args.adapter,
        connect_timeout=args.connect_timeout,
        history_read_timeout=args.read_timeout,
    ) as client:
        history = await client.history()
        if args.output == "text":
            print("#\tBq/m3\tpCi/L")
            for index, value_bq_m3 in enumerate(history["values_bq_m3"]):
                value_pci_l = history["values_pci_l"][index]
                print(f"{index + 1}\t{value_bq_m3}\t{value_pci_l}")
        else:
            print(json.dumps(history, separators=(",", ":")))


async def main():
    parser = ArgumentParser(
        description="Ecosense RadonEye command line interface (currently supports RD200 v1/v2)",
        formatter_class=ArgumentDefaultsHelpFormatter,
        prog="radoneye",
    )

    parser.add_argument("--adapter", type=int, help="bluetooth adapter name (hci0 on Linux)")

    subparsers = parser.add_subparsers(required=True, help="sub-command help")

    parser_list = subparsers.add_parser(
        "list",
        help="scan for devices available nearby",
        formatter_class=ArgumentDefaultsHelpFormatter,
    )
    parser_list.add_argument("--timeout", type=int, help="scan timeout", default=5)
    parser_list.add_argument(
        "--output", choices=["json", "text"], help="output format", default="text"
    )
    parser_list.set_defaults(func=cmd_list)

    parser_beep = subparsers.add_parser(
        "beep",
        help="trigger beep",
        formatter_class=ArgumentDefaultsHelpFormatter,
    )
    parser_beep.add_argument("--connect-timeout", type=int, help="connect timeout", default=10)
    parser_beep.add_argument("address", help="device address")
    parser_beep.set_defaults(func=cmd_beep)

    parser_status = subparsers.add_parser(
        "status",
        help="read status",
        formatter_class=ArgumentDefaultsHelpFormatter,
    )
    parser_status.add_argument("--connect-timeout", type=int, help="connect timeout", default=10)
    parser_status.add_argument("--read-timeout", type=int, help="read timeout", default=5)
    parser_status.add_argument(
        "--output", choices=["json", "text"], help="output format", default="text"
    )
    parser_status.add_argument("address", help="device address")
    parser_status.set_defaults(func=cmd_status)

    parser_history = subparsers.add_parser(
        "history",
        help="read history",
        formatter_class=ArgumentDefaultsHelpFormatter,
    )
    parser_history.add_argument("--connect-timeout", type=int, help="connect timeout", default=10)
    parser_history.add_argument("--read-timeout", type=int, help="read timeout", default=60)
    parser_history.add_argument(
        "--output", choices=["json", "text"], help="output format", default="text"
    )
    parser_history.add_argument("address", help="device address")
    parser_history.set_defaults(func=cmd_history)

    args = parser.parse_args()
    await args.func(args)


if __name__ == "__main__":
    asyncio.run(main())



import asyncio
import math
from struct import unpack

from bleak import BleakClient
from bleak.backends.characteristic import BleakGATTCharacteristic

from radoneye.interface import RadonEyeHistory, RadonEyeInterfaceBase, RadonEyeStatus

SERVICE_V1 = "00001523-1212-efde-1523-785feabcd123"
CHARACTERISTIC_V1_COMMAND = "00001524-1212-efde-1523-785feabcd123"
CHARACTERISTIC_V1_STATUS = "00001525-1212-efde-1523-785feabcd123"
COMMAND_V1_STATUS1 = [0x50]
COMMAND_V1_STATUS2 = [0x51]


class RadonEyeMessageParserV1:
    @classmethod
    def parse_status(cls, data: bytearray, data2: bytearray, addr: str) -> RadonEyeStatus:
        latest_value = float(unpack("<f", data[2:6])[0])
        latest_pci_l = round(latest_value, 2)
        latest_bq_m3 = round(latest_value * 37)

        day_avg_value = float(unpack("<f", data[6:10])[0])
        day_avg_pci_l = round(day_avg_value, 2)
        day_avg_bq_m3 = round(day_avg_value * 37)

        month_avg_value = float(unpack("<f", data[10:14])[0])
        month_avg_pci_l = round(month_avg_value, 2)
        month_avg_bq_m3 = round(month_avg_value * 37)

        peak_value = float(unpack("<f", data2[12:16])[0])
        peak_pci_l = round(peak_value, 2)
        peak_bq_m3 = round(peak_value * 37)

        uptime_minutes = unpack("<I", data2[4:8])[0]
        uptime_days = math.floor(uptime_minutes / (60 * 24))
        uptime_hours = math.floor(uptime_minutes % (60 * 24) / 60)
        uptime_mins = uptime_minutes % 60
        uptime_str = f"{uptime_days}d{uptime_hours:02}h{uptime_mins:02}m"

        return {
            # serial is needed since it is used as unique device id
            # in absense of true serial, we have to use device address
            "serial": addr.replace(":", "").replace("-", "").upper(),
            "model": "RD200",
            "version": "?",
            "latest_bq_m3": latest_bq_m3,
            "latest_pci_l": latest_pci_l,
            "day_avg_bq_m3": day_avg_bq_m3,
            "day_avg_pci_l": day_avg_pci_l,
            "month_avg_bq_m3": month_avg_bq_m3,
            "month_avg_pci_l": month_avg_pci_l,
            "peak_bq_m3": peak_bq_m3,
            "peak_pci_l": peak_pci_l,
            "counts_current": 0,
            "counts_previous": 0,
            "counts_str": "?/?",
            "uptime_minutes": uptime_minutes,
            "uptime_str": uptime_str,
        }


class RadonEyeAdapterV1(RadonEyeInterfaceBase):
    status_read_timeout: float

    def __init__(self, status_read_timeout: float) -> None:
        self.status_read_timeout = status_read_timeout

    @classmethod
    def supports(cls, client: BleakClient) -> bool:
        return bool(client.services.get_service(SERVICE_V1))

    async def status(self, client: BleakClient) -> RadonEyeStatus:
        loop = asyncio.get_running_loop()
        future = loop.create_future()
        data1: bytearray | None = None
        data2: bytearray | None = None

        def callback(char: BleakGATTCharacteristic, data: bytearray) -> None:
            nonlocal data1
            nonlocal data2
            if data[0] == COMMAND_V1_STATUS1[0]:
                data1 = data
            elif data[0] == COMMAND_V1_STATUS2[0]:
                data2 = data
            if data1 and data2:
                status = RadonEyeMessageParserV1.parse_status(data1, data2, client.address)
                future.set_result(status)

        await client.start_notify(CHARACTERISTIC_V1_STATUS, callback)  # type: ignore
        await client.write_gatt_char(CHARACTERISTIC_V1_COMMAND, bytearray(COMMAND_V1_STATUS1))
        await client.write_gatt_char(CHARACTERISTIC_V1_COMMAND, bytearray(COMMAND_V1_STATUS2))
        result = await asyncio.wait_for(future, timeout=self.status_read_timeout)
        await client.stop_notify(CHARACTERISTIC_V1_STATUS)
        return result

    async def history(self, client: BleakClient) -> RadonEyeHistory:
        raise NotImplementedError

    async def beep(self, client: BleakClient) -> None:
        raise NotImplementedError



import asyncio
import math
from struct import unpack
from typing import cast

from bleak import BleakClient
from bleak.backends.characteristic import BleakGATTCharacteristic

from radoneye.interface import (
    RadonEyeHistory,
    RadonEyeHistoryPage,
    RadonEyeInterfaceBase,
    RadonEyeStatus,
)

SERVICE_V2 = "00001523-0000-1000-8000-00805f9b34fb"
CHARACTERISTIC_V2_COMMAND = "00001524-0000-1000-8000-00805f9b34fb"
CHARACTERISTIC_V2_STATUS = "00001525-0000-1000-8000-00805f9b34fb"
CHARACTERISTIC_V2_HISTORY = "00001526-0000-1000-8000-00805f9b34fb"
COMMAND_V2_STATUS = [0x40]
COMMAND_V2_HISTORY = [0x41]
COMMAND_V2_BEEP = [0xA1, 0x11, 0x17, 0x0C, 0x0B, 0x01, 0x25, 0x28]

BEEP_DELAY = 0.2  # sec


class RadonEyeMessageParserV2:
    @classmethod
    def read_short(cls, data: bytearray, start: int) -> int:
        return unpack("<H", data[slice(start, start + 2)])[0]

    @classmethod
    def read_str(cls, data: bytearray, start: int, length: int) -> str:
        return data[slice(start, start + length)].decode()

    @classmethod
    def to_pci_l(cls, value_bq_m3: int) -> float:
        return round(value_bq_m3 / 37, 2)

    @classmethod
    def parse_status(cls, data: bytearray) -> RadonEyeStatus:
        serial = cls.read_str(data, 8, 3) + cls.read_str(data, 2, 6) + cls.read_str(data, 11, 4)
        model = cls.read_str(data, 16, 6)
        version = cls.read_str(data, 22, 6)
        latest_bq_m3 = cls.read_short(data, 33)
        latest_pci_l = cls.to_pci_l(latest_bq_m3)
        day_avg_bq_m3 = cls.read_short(data, 35)
        day_avg_pci_l = cls.to_pci_l(day_avg_bq_m3)
        month_avg_bq_m3 = cls.read_short(data, 37)
        month_avg_pci_l = cls.to_pci_l(month_avg_bq_m3)
        counts_current = cls.read_short(data, 39)
        counts_previous = cls.read_short(data, 41)
        counts_str = f"{counts_current}/{counts_previous}"
        uptime_minutes = cls.read_short(data, 43)
        uptime_days = math.floor(uptime_minutes / (60 * 24))
        uptime_hours = math.floor(uptime_minutes % (60 * 24) / 60)
        uptime_mins = uptime_minutes % 60
        uptime_str = f"{uptime_days}d{uptime_hours:02}h{uptime_mins:02}m"
        peak_bq_m3 = cls.read_short(data, 51)
        peak_pci_l = cls.to_pci_l(peak_bq_m3)

        return {
            "serial": serial,
            "model": model,
            "version": version,
            "latest_bq_m3": latest_bq_m3,
            "latest_pci_l": latest_pci_l,
            "day_avg_bq_m3": day_avg_bq_m3,
            "day_avg_pci_l": day_avg_pci_l,
            "month_avg_bq_m3": month_avg_bq_m3,
            "month_avg_pci_l": month_avg_pci_l,
            "peak_bq_m3": peak_bq_m3,
            "peak_pci_l": peak_pci_l,
            "counts_current": counts_current,
            "counts_previous": counts_previous,
            "counts_str": counts_str,
            "uptime_minutes": uptime_minutes,
            "uptime_str": uptime_str,
        }

    @classmethod
    def parse_history_page(cls, data: bytearray) -> RadonEyeHistoryPage:
        data.pop(0)  # command

        page_count = data.pop(0)
        page_no = data.pop(0)
        value_count = data.pop(0)
        values_bq_m3 = cast(list[int], unpack("<" + "H" * (len(data) // 2), data))
        values_pci_l = [cls.to_pci_l(x) for x in values_bq_m3]

        return {
            "page_count": page_count,
            "page_no": page_no,
            "value_count": value_count,
            "values_bq_m3": values_bq_m3,
            "values_pci_l": values_pci_l,
        }

    @classmethod
    def merge_history(cls, pages: list[RadonEyeHistoryPage]) -> RadonEyeHistory:
        pages = sorted(pages, key=lambda page: page["page_no"])
        if pages:
            for index, page in enumerate(pages):
                if index + 1 != page["page_no"]:
                    raise ValueError("History page order mismatch")
            if pages[0]["page_count"] != len(pages):
                raise ValueError("History page count mismatch")
        return {
            "values_bq_m3": [value for page in pages for value in page["values_bq_m3"]],
            "values_pci_l": [value for page in pages for value in page["values_pci_l"]],
        }


class RadonEyeAdapterV2(RadonEyeInterfaceBase):
    status_read_timeout: float
    history_read_timeout: float

    def __init__(self, status_read_timeout: float, history_read_timeout: float) -> None:
        self.status_read_timeout = status_read_timeout
        self.history_read_timeout = history_read_timeout

    @classmethod
    def supports(cls, client: BleakClient) -> bool:
        return bool(client.services.get_service(SERVICE_V2))

    async def status(self, client: BleakClient) -> RadonEyeStatus:
        loop = asyncio.get_running_loop()
        future = loop.create_future()

        def callback(char: BleakGATTCharacteristic, data: bytearray) -> None:
            if data[0] == COMMAND_V2_STATUS[0]:
                future.set_result(RadonEyeMessageParserV2.parse_status(data))

        await client.start_notify(CHARACTERISTIC_V2_STATUS, callback)  # type: ignore
        await client.write_gatt_char(CHARACTERISTIC_V2_COMMAND, bytearray(COMMAND_V2_STATUS))
        result = await asyncio.wait_for(future, timeout=self.status_read_timeout)
        await client.stop_notify(CHARACTERISTIC_V2_STATUS)
        return result

    async def history(self, client: BleakClient) -> RadonEyeHistory:
        loop = asyncio.get_running_loop()
        future = loop.create_future()
        pages: list[RadonEyeHistoryPage] = []

        def callback(char: BleakGATTCharacteristic, data: bytearray) -> None:
            if data[0] == COMMAND_V2_HISTORY[0]:
                page = RadonEyeMessageParserV2.parse_history_page(data)
                pages.append(page)
                if page["page_count"] == page["page_no"]:
                    future.set_result(RadonEyeMessageParserV2.merge_history(pages))

        await client.start_notify(CHARACTERISTIC_V2_HISTORY, callback)  # type: ignore
        await client.write_gatt_char(CHARACTERISTIC_V2_COMMAND, bytearray(COMMAND_V2_HISTORY))
        result = await asyncio.wait_for(future, timeout=self.history_read_timeout)
        await client.stop_notify(CHARACTERISTIC_V2_HISTORY)
        return result

    async def beep(self, client: BleakClient) -> None:
        await client.write_gatt_char(CHARACTERISTIC_V2_COMMAND, bytearray(COMMAND_V2_BEEP))
        # there is some delay needed before you can do next beep, otherwise it will be just one beep
        await asyncio.sleep(BEEP_DELAY)



from typing import Union

from bleak import BleakClient
from bleak.backends.device import BLEDevice

from radoneye.interface import RadonEyeHistory, RadonEyeInterfaceBase, RadonEyeStatus
from radoneye.interface_factory import RadonEyeInterfaceFactory


class RadonEyeClient:
    client: BleakClient
    api: RadonEyeInterfaceBase

    adapter: str | None
    connect_timeout: float
    status_read_timeout: float
    history_read_timeout: float

    def __init__(
        self,
        address_or_ble_device: Union[BLEDevice, str],
        connect_timeout: float = 10,
        status_read_timeout: float = 5,
        history_read_timeout: float = 60,
        adapter: str | None = None,
    ) -> None:
        self.client = BleakClient(address_or_ble_device, timout=connect_timeout, adapter=adapter)
        self.adapter = adapter
        self.connect_timeout = connect_timeout
        self.status_read_timeout = status_read_timeout
        self.history_read_timeout = history_read_timeout

    def __init_api(self) -> None:
        self.api = RadonEyeInterfaceFactory.create(
            self.client,
            status_read_timeout=self.status_read_timeout,
            history_read_timeout=self.history_read_timeout,
        )

    async def __aenter__(self):
        await self.client.connect()  # type: ignore
        self.__init_api()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):  # type: ignore
        await self.client.disconnect()

    async def connect(self) -> None:
        await self.client.connect()  # type: ignore
        self.__init_api()

    async def disconnect(self) -> None:
        await self.client.disconnect()  # type: ignore

    @property
    def is_connected(self) -> bool:
        return self.client.is_connected

    async def beep(self) -> None:
        await self.api.beep(self.client)

    async def status(self) -> RadonEyeStatus:
        return await self.api.status(self.client)

    async def history(self) -> RadonEyeHistory:
        return await self.api.history(self.client)
